#include <winsock2.h>
#include <windows.h>
#include <tchar.h>
#include <stdio.h>
#include <errno.h>
#include <crtdbg.h>
#include <OCIdl.h>
#include <Olectl.h>

#include "sresource.h"


#pragma comment(lib, "user32")
#pragma comment(lib, "ws2_32")
#pragma comment(lib, "gdi32")
#pragma comment(lib, "ole32")
#pragma comment(lib, "oleaut32")

typedef struct Init
{
	HRESULT hr;
    WSADATA wsaData;
    int     res;

	Init()
	{
		hr = CoInitialize(NULL);
	    res = WSAStartup(MAKEWORD(2,2), &wsaData);
	}

	~Init()
	{
	    WSACleanup();
		if (SUCCEEDED(hr))
			CoUninitialize();
	}

	BOOL isOkay()
	{
	    return (res == NO_ERROR) && SUCCEEDED(hr);
	}
} Init;

typedef struct StreamData
{
	HGLOBAL hGlobal;
	LPBYTE  data;
	ULONG   size;
	StreamData(IStream* pStream)
	{
		hGlobal = NULL;
		HRESULT hr = GetHGlobalFromStream(pStream, &hGlobal);
		_ASSERT(SUCCEEDED(hr));
		size = (ULONG)GlobalSize(hGlobal);
		data = (LPBYTE)GlobalLock(hGlobal);
	}
	~StreamData() 
	{
		GlobalUnlock(hGlobal);
	}
} StreamData;

typedef struct RxSOCKET
{
    int res;
	int wsaErr;
    SOCKET sock, rsock;
    struct sockaddr_in srv_in; 
	static const int nPort = 15777;

	RxSOCKET(): 
		sock(INVALID_SOCKET),
		rsock(INVALID_SOCKET),
		res(SOCKET_ERROR), 
		wsaErr(0)
	{
	}
	~RxSOCKET() { close(); }

	SOCKET listen()
	{
		sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if (sock == INVALID_SOCKET) {
			close();
			return false;
		}

		sockaddr_in srv;
		srv.sin_family = AF_INET;
		srv.sin_addr.s_addr = inet_addr("0.0.0.0");
		srv.sin_port = htons( nPort );

		if (SOCKET_ERROR == ::bind( sock, (SOCKADDR*) &srv, sizeof(srv))) 
		{
			close();
			return false;
		}

		if (SOCKET_ERROR == ::listen( sock, SOMAXCONN ))
		{
			close();
			return false;
		}

		int rsize = sizeof(srv);
		rsock = ::accept(sock, (SOCKADDR*) &srv, &rsize);
		if (INVALID_SOCKET == rsock)
		{
			close();
			return false;
		}

		return (rsock != INVALID_SOCKET);
	}

	void close()
	{
		if (rsock != INVALID_SOCKET)
		{
			closesocket(rsock);
			rsock = INVALID_SOCKET;
		}

		if (sock != INVALID_SOCKET)
		{
			closesocket(sock);
			sock = INVALID_SOCKET;
		}
	}

	int recvToStream(IStream** pStream)
	{
		if (*pStream)
		{
			(*pStream)->Release();
			*pStream = NULL;
		}

		HRESULT hr = ::CreateStreamOnHGlobal(NULL, TRUE, pStream);
		if (FAILED(hr))
		{
			close();
			return -1;
		}

		char rbuf[1024];
		int rbuflen = sizeof(rbuf);
		size_t total = 0;
		do 
		{
			res = ::recv(rsock, rbuf, rbuflen, 0);
			if ( res > 0 )
			{
				ULONG written;
				(*pStream)->Write(rbuf, res, &written);
				total += res;
			}
			else
				wsaErr = WSAGetLastError();

		} 
		while( res > 0 );

		char sbuf[] = "OK";
	    res = ::send( rsock, (LPCSTR)sbuf, sizeof(sbuf), 0 );
	    if (res == SOCKET_ERROR) 
	    {
	        close();
	    }

		return res;
	}
} RxSOCKET;

Init init;
HBITMAP hbm = NULL;
IStream* pStream = NULL;
IPicture* pPict = NULL;


void Draw(HDC dc)
{
	BITMAP bm = {0};
	if (hbm && GetObject(hbm, sizeof(bm), &bm))
	{
		HDC cdc = CreateCompatibleDC(dc);
		HGDIOBJ oldBm = SelectObject(cdc, (HGDIOBJ)hbm);
		BitBlt(dc, 0, 0, bm.bmWidth, bm.bmHeight, cdc, 0, 0, SRCCOPY);
		SelectObject(cdc, oldBm);
		DeleteDC(cdc);
	}
}

LONG Save()
{
	if (pPict)
		pPict->Release();

	LONG pictSize = 0;
	HRESULT hr = S_OK;
	if (pStream)
	{
		StreamData image(pStream);
		pictSize = image.size;

#ifdef DEBUG
		FILE* f = fopen("server.bmp", "w");
		if (f)
		{
			fwrite(image.data, 1, image.size, f);
		}
		fclose(f);
#endif
	}
	else
		return -1;
		
	LARGE_INTEGER offs = {0};
	hr = pStream->Seek(offs, STREAM_SEEK_SET, NULL);
	hr = ::OleLoadPicture(pStream, pictSize, FALSE, IID_IPicture, (void**)&pPict);
	if (SUCCEEDED(hr))
	{
		_ASSERT(pPict);
#if 1
		if (pStream)
		{
			pStream->Release();
			pStream = NULL;
		}
#endif
		hr = pPict->get_Handle((OLE_HANDLE*)&hbm);
	}
	else
	{
		_ASSERT(FALSE);
	}

	return pictSize;
}

DWORD CALLBACK Listener(LPVOID pVoid)
{
	HWND hDlg = (HWND)pVoid;
	do
	{
		RxSOCKET rx;
		rx.listen();
		rx.recvToStream(&pStream);
		Save();
		InvalidateRect(hDlg, NULL, TRUE);
	}
	while(1);

	return 0;
}

BOOL OnInitDialog(HWND hDlg, WPARAM wp, LPARAM lp)
{
	DWORD threadId;
	CloseHandle(CreateThread(NULL, 0, Listener, (LPVOID)hDlg, 0, &threadId));
	return TRUE;
}

void OnPaint(HWND hDlg)
{
	PAINTSTRUCT ps;
	HDC dc = BeginPaint(hDlg, &ps);
	Draw(dc);
	EndPaint(hDlg, &ps);
}


BOOL CALLBACK DlgProc(HWND hDlg, UINT msg, WPARAM wp, LPARAM lp)
{
	switch (msg)
	{
	case WM_INITDIALOG:
		return OnInitDialog(hDlg, wp, lp);

	case WM_PAINT:
		OnPaint(hDlg);
		break;

	case WM_COMMAND:
		switch (LOWORD(wp))
		{
		case IDCANCEL:
			EndDialog(hDlg, 0);
			break;
		}
		break;
	}
	return 0;
}

int WINAPI _tWinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmd, INT nShow)
{
	if (!init.isOkay())
	{
		TCHAR errMsg[512] = {0};
		wsprintf(errMsg, TEXT("res = %d, hr = 0x%08x"), init.res, init.hr);
		MessageBox(NULL, errMsg, TEXT("Initialization failure"), MB_OK | MB_ICONHAND);
		return 1;
	}

	int res = DialogBox(hInst, MAKEINTRESOURCE(IDD_MAIN), NULL, (DLGPROC)DlgProc);

	return 0;
}
